import argparse
from datetime import datetime
import re
import sys
from typing import Optional, Sequence, Union

from jira import Issue, JIRA
from jira.resources import Resource, User, Version

from rhjira import util


def getnames(fieldlist: Optional[Sequence[Union[Resource, User, Version]]]) -> str:
    retstr = ""
    count = 0
    if fieldlist:
        for entry in fieldlist:
            count += 1
            retstr = retstr + f"{entry.name}"
            if count != len(fieldlist):
                retstr = retstr + ", "
    return retstr


def defaultShowText(jira: JIRA, issue: Issue) -> str:
    epicname = ""
    epiclink = ""
    parentlink = ""

    if issue.fields.issuetype is not None:
        if issue.fields.issuetype.name == "Epic":
            epicname = issue.fields.customfield_12311141
            parentlink = issue.fields.customfield_12313140
            parentlink = "" if parentlink is None else parentlink
        else:
            epiclink = issue.fields.customfield_12311140
            epiclink = "" if epiclink is None else epiclink

    components = getnames(issue.fields.components)
    affectsversions = getnames(issue.fields.versions)
    fixversions = getnames(issue.fields.fixVersions)
    contributors = getnames(issue.fields.customfield_12315950)

    releaseblocker = ""
    if hasattr(issue.fields, 'customfield_12319743') and issue.fields.customfield_12319743:
        releaseblocker = issue.fields.customfield_12319743.value

    severity = ""
    if issue.fields.customfield_12316142:
        severity = issue.fields.customfield_12316142.value

    gitpullrequest = ""
    if issue.fields.customfield_12310220:
        for url in issue.fields.customfield_12310220:
            gitpullrequest = url
            break

    assignee = ""
    if issue.fields.assignee is not None:
        assignee = issue.fields.assignee.name

    # Get blocking relationships
    blocks = []
    isblockedby = []
    if hasattr(issue.fields, 'issuelinks'):
        for link in issue.fields.issuelinks:
            if hasattr(link, 'outwardIssue'):
                summary = link.outwardIssue.raw.get('fields', {}).get('summary', '')
                status = link.outwardIssue.raw.get('fields', {}).get('status', {}).get('name', '')
                status_category = link.outwardIssue.raw.get('fields', {}).get('status', {}).get('statusCategory', {}).get('name', '')
                if link.type.outward == 'blocks':
                    blocks.append((link.outwardIssue.key, summary, status, status_category))
            if hasattr(link, 'inwardIssue'):
                summary = link.inwardIssue.raw.get('fields', {}).get('summary', '')
                status = link.inwardIssue.raw.get('fields', {}).get('status', {}).get('name', '')
                status_category = link.inwardIssue.raw.get('fields', {}).get('status', {}).get('statusCategory', {}).get('name', '')
                if link.type.inward == 'is blocked by':
                    isblockedby.append((link.inwardIssue.key, summary, status, status_category))

    # Initialize child_issues
    child_issues = []
    # Get child issues for Features and Epics
    if issue.fields.issuetype.name in ["Feature", "Epic"]:
        child_issues = util.searchissues(jira, f"'Parent Link' = {issue.key}", 2000)

    # Get Epic Link and Parent Link summaries
    epiclink_summary = ""
    epiclink_status = ""
    epiclink_category = ""
    parentlink_summary = ""
    parentlink_status = ""
    parentlink_category = ""
    if epiclink:
        try:
            epic = util.getissue(jira, epiclink)
            epiclink_summary = epic.fields.summary
            epiclink_status = epic.fields.status.name
            epiclink_category = epic.fields.status.statusCategory.name
        except Exception:
            pass
    if parentlink:
        try:
            parent = util.getissue(jira, parentlink)
            parentlink_summary = parent.fields.summary
            parentlink_status = parent.fields.status.name
            parentlink_category = parent.fields.status.statusCategory.name
        except Exception:
            pass

    # Build the output lines
    lines = [
        f"{issue.key}: {issue.fields.summary}",
        "===================================",
        f"{issue.fields.description}",
        "===================================",
        f"Epic Name: {epicname}",
        f"Ticket Type: {issue.fields.issuetype}",
        f"Status: {issue.fields.status}",
        f"Creator: {issue.fields.creator.name}",
        f"Assignee: {assignee}",
        f"Components: {components}",
        f"Affects Versions: {affectsversions}",
        f"Fix Versions: {fixversions}",
        f"Priority: {issue.fields.priority.name}",
        f"Contributors: {contributors}",
        f"Release Blocker: {releaseblocker}",
        f"Severity: {severity}",
        f"Git Pull Request: {gitpullrequest}",
    ]

    # Add Epic Link if it exists
    if epiclink:
        lines.append(f"Epic Link: {util.format_link(epiclink, epiclink_summary, epiclink_status, epiclink_category)}")

    # Add Parent Link if it exists
    if parentlink:
        lines.append(f"Parent Link: {util.format_link(parentlink, parentlink_summary, parentlink_status, parentlink_category)}")

    # Add Blocks if they exist
    if blocks:
        lines.extend(f"Blocks: {util.format_link(key, summary, status, status_category)}" for key, summary, status, status_category in blocks)

    # Add Is Blocked By if they exist
    if isblockedby:
        lines.extend(f"Is Blocked By: {util.format_link(key, summary, status, status_category)}" for key, summary, status, status_category in isblockedby)

    # Add Child Issues if they exist
    if child_issues:
        lines.extend(f"Child Issue: {util.format_link(child.key, child.fields.summary, child.fields.status.name, child.fields.status.statusCategory.name)}" for child in child_issues)

    # Filter out empty lines and join with newlines
    return "\n".join(line for line in lines if line.strip())


def show(jira: JIRA) -> None:
    # handle arguments
    sys.argv.remove("show")
    parser = argparse.ArgumentParser(
        description="Show basic information on a RH jira ticket"
    )
    args, ticketIDs = parser.parse_known_args()

    if len(ticketIDs) != 1:
        print(f"Error: ticketID not clear or found: {ticketIDs}")
        sys.exit(1)
    ticketID = ticketIDs[0]

    try:
        issue = util.getissue(jira, ticketID)
    except Exception as e:
        util.handle_jira_error(e, f"lookup ticket {ticketID}")
        sys.exit(1)

    outtext = defaultShowText(jira, issue)
    if issue.fields.issuetype.name == "Epic":
        outtext = re.sub(r"^Epic Link:.*\n?", "", outtext, flags=re.MULTILINE)
    else:
        outtext = re.sub(r"^Epic Name:.*\n?", "", outtext, flags=re.MULTILINE)
        outtext = re.sub(r"^Parent Link:.*\n?", "", outtext, flags=re.MULTILINE)

    if issue.fields.issuetype.name not in ["Bug", "Story"]:
        outtext = re.sub(r"^Git Pull Request:.*\n?", "", outtext, flags=re.MULTILINE)

    print(outtext)

    if not issue.fields.comment:
        return

    numcomments = len(issue.fields.comment.comments)
    print(f"--------- {numcomments}  Comments ---------")

    count = 0
    for comment in issue.fields.comment.comments:
        count += 1
        created = datetime.strptime(comment.created, "%Y-%m-%dT%H:%M:%S.%f%z")
        print(f"Comment #{count} | {created.strftime('%c')} | {comment.author.name}")
        print("")
        print(f"{comment.body}")
        print("")



from unittest.mock import MagicMock, patch

from jira import Issue, JIRA
import pytest

from rhjira.show import defaultShowText


@pytest.fixture(autouse=False)
def mock_jira() -> JIRA:
    return MagicMock(spec=JIRA)

@pytest.fixture(autouse=False)
def mock_issue() -> Issue:
    issue = MagicMock(spec=Issue)
    issue.key = "RHEL-1234"

    # Create fields as a MagicMock
    fields = MagicMock()
    fields.summary = "Test Issue"
    fields.description = "Test Description"
    fields.issuetype = MagicMock()
    fields.issuetype.name = "Bug"
    fields.status = MagicMock()
    fields.status.name = "In Progress"
    fields.creator = MagicMock()
    fields.creator.name = "testuser"
    fields.assignee = None
    fields.components = []
    fields.versions = []
    fields.fixVersions = []
    fields.priority = MagicMock()
    fields.priority.name = "High"
    fields.customfield_12315950 = []  # Contributors
    fields.customfield_12319743 = None  # Release Blocker
    fields.customfield_12316142 = None  # Severity
    fields.customfield_12310220 = []  # Git Pull Request
    fields.issuelinks = []
    fields.comment = None

    issue.fields = fields
    return issue

@patch('rhjira.show.util.getissue')
def test_format_link_with_epic(mock_getissue: MagicMock,
                               mock_jira: JIRA,
                               mock_issue: Issue) -> None:
    # Setup epic link
    mock_issue.fields.customfield_12311140 = "RHEL-5678"

    # Mock epic issue
    epic = MagicMock(spec=Issue)
    epic.fields = MagicMock()
    epic.fields.summary = "Epic Summary"
    epic.fields.status = MagicMock()
    epic.fields.status.name = "In Progress"
    epic.fields.status.statusCategory = MagicMock()
    epic.fields.status.statusCategory.name = "In Progress"
    mock_getissue.return_value = epic

    output = defaultShowText(mock_jira, mock_issue)
    assert "Epic Link:" in output
    assert "RHEL-5678" in output
    assert "Epic Summary" in output
    assert "In Progress" in output

@patch('rhjira.show.util.getissue')
def test_format_link_with_parent(mock_getissue: MagicMock,
                                 mock_jira: JIRA,
                                 mock_issue: Issue) -> None:
    # Setup parent link for Epic type
    mock_issue.fields.issuetype.name = "Epic"  # type: ignore[attr-defined]
    mock_issue.fields.customfield_12313140 = "RHEL-5678"
    mock_issue.fields.customfield_12311141 = "Epic Name"  # Epic name field

    # Mock parent issue
    parent = MagicMock(spec=Issue)
    parent.fields = MagicMock()
    parent.fields.summary = "Parent Summary"
    parent.fields.status = MagicMock()
    parent.fields.status.name = "To Do"
    parent.fields.status.statusCategory = MagicMock()
    parent.fields.status.statusCategory.name = "To Do"
    mock_getissue.return_value = parent

    output = defaultShowText(mock_jira, mock_issue)
    assert "Parent Link:" in output
    assert "RHEL-5678" in output
    assert "Parent Summary" in output
    assert "To Do" in output

def test_format_link_with_blocks(mock_jira: JIRA, mock_issue: Issue) -> None:
    # Setup blocking link
    block_issue = MagicMock()
    block_issue.key = "RHEL-5678"
    block_issue.raw = {
        "fields": {
            "summary": "Blocking Issue",
            "status": {
                "name": "Done",
                "statusCategory": {"name": "Done"}
            }
        }
    }
    link = MagicMock()
    link.type.outward = "blocks"
    link.outwardIssue = block_issue
    mock_issue.fields.issuelinks = [link]

    output = defaultShowText(mock_jira, mock_issue)
    assert "Blocks:" in output
    assert "RHEL-5678" in output
    assert "Blocking Issue" in output
    assert "Done" in output

def test_format_link_with_is_blocked_by(mock_jira: JIRA, mock_issue: Issue) -> None:
    # Setup blocked by link
    blocker_issue = MagicMock()
    blocker_issue.key = "RHEL-5678"
    blocker_issue.raw = {
        "fields": {
            "summary": "Blocker Issue",
            "status": {
                "name": "In Progress",
                "statusCategory": {"name": "In Progress"}
            }
        }
    }
    link = MagicMock()
    link.type.inward = "is blocked by"
    link.inwardIssue = blocker_issue
    mock_issue.fields.issuelinks = [link]

    output = defaultShowText(mock_jira, mock_issue)
    assert "Is Blocked By:" in output
    assert "RHEL-5678" in output
    assert "Blocker Issue" in output
    assert "In Progress" in output

@patch('rhjira.show.util.searchissues')
def test_format_link_with_child_issues(mock_searchissues: MagicMock,
                                       mock_jira: JIRA,
                                       mock_issue: Issue) -> None:
    # Setup child issues
    mock_issue.fields.issuetype.name = "Feature"  # type: ignore[attr-defined]
    child = MagicMock(spec=Issue)
    child.key = "RHEL-5678"
    child.fields = MagicMock()
    child.fields.summary = "Child Issue"
    child.fields.status = MagicMock()
    child.fields.status.name = "To Do"
    child.fields.status.statusCategory = MagicMock()
    child.fields.status.statusCategory.name = "To Do"
    mock_searchissues.return_value = [child]

    output = defaultShowText(mock_jira, mock_issue)
    assert "Child Issue:" in output
    assert "RHEL-5678" in output
    assert "Child Issue" in output
    assert "To Do" in output

@patch('rhjira.show.util.getissue')
def test_format_link_without_links(mock_getissue: MagicMock,
                                   mock_jira: JIRA,
                                   mock_issue: Issue) -> None:
    # Ensure no links are set
    mock_issue.fields.customfield_12311140 = None
    mock_issue.fields.customfield_12313140 = None
    mock_issue.fields.issuelinks = []
    mock_issue.fields.issuetype.name = "Bug"  # type: ignore[attr-defined]

    output = defaultShowText(mock_jira, mock_issue)
    assert "Epic Link:" not in output
    assert "Parent Link:" not in output
    assert "Blocks:" not in output
    assert "Is Blocked By:" not in output
    assert "Child Issue:" not in output



"""Red Hat Jira CLI tool."""

from .cli import main
from .comment import comment
from .create import create
from .dump import dump
from .edit import edit
from .hierarchy import hierarchy
from .info import info
from .list import list
from .login import login, setpassword
from .show import show

__all__ = [
    'comment',
    'create',
    'dump',
    'edit',
    'hierarchy',
    'info',
    'list',
    'login',
    'main',
    'setpassword',
    'show',
]



import sys
from typing import NoReturn

from jira import JIRAError

import rhjira
from rhjira import util


def usage() -> NoReturn:
    print("Usage:")
    print("  rhjira [flags] [command]")
    print(" ")
    print("Available Commands:")
    print("  comment     Comment on a jira ticket on https://issues.redhat.com")
    print("  create      Create a jira ticket on https://issues.redhat.com")
    print("  dump        Dump jira issue variables.")
    print("  edit        Edit a jira ticket on https://issues.redhat.com")
    print("  help        Help about any command")
    print("  info        Display project information (components and versions)")
    print("  list        List issues")
    print(
        "  settoken    Save a jira token to the keyring (more secure than using $JIRA_TOKEN)"
    )
    print("  show        Show a jira ticket on https://issues.redhat.com")
    print("  hierarchy   Show hierarchy information for a jira ticket")
    print(" ")
    print("For help on individual commands, execute rhjira [command] --help")
    sys.exit(1)


def main() -> None:
    if len(sys.argv) <= 1:
        usage()

    if sys.argv[1] != "settoken":
        try:
            jira = rhjira.login()
        except JIRAError as e:
            util.handle_jira_error(e, "login")
            sys.exit(1)

    match sys.argv[1]:
        case "comment":
            rhjira.comment(jira)
        case "close":
            rhjira.edit(jira)
        case "create":
            rhjira.create(jira)
        case "dump":
            rhjira.dump(jira)
        case "edit":
            rhjira.edit(jira)
        case "info":
            rhjira.info(jira)
        case "list":
            rhjira.list(jira)
        case "settoken":
            rhjira.setpassword()
        case "show":
            rhjira.show(jira)
        case "hierarchy":
            rhjira.hierarchy(jira)
        case _:
            usage()


if __name__ == "__main__":
    main()



from datetime import datetime
import json
import os
import re
import shlex
import subprocess
import sys
import time
from typing import (
    Any,
    Callable,
    cast,
    Dict,
    List,
    Optional,
    overload,
    Protocol,
    Sequence,
    TypeVar,
    Union,
)

from jira import Issue, JIRA, JIRAError
from jira.resources import (
    Attachment,
    Comment,
    Component,
    Group,
    IssueLink,
    Resource,
    SecurityLevel,
    User,
    Version,
    Votes,
    Watchers,
)

from rhjira import login


# Custom type definitions for missing Jira types
class BugzillaBug(Protocol):
    bugid: str


class JiraOption(Protocol):
    value: str
    child: Optional["JiraOption"]


class JiraProgress(Protocol):
    progress: int


class JiraNamedResource(Protocol):
    name: str


# Type variables for generic functions
T = TypeVar("T", bound=Resource)
StrConvertible = Union[str, int, float, bool, None]

MAX_RETRIES = 6


def convertdescription(description: str) -> str:
    return description.replace(r"\n", "\n")


def dumpissue(issue: Issue) -> None:
    print(json.dumps(issue.raw, indent=4))


def rhjiratax() -> None:
    # RH's Jira instance has a 2/second/node rate limit.  To avoid
    # this the code has to implement a half second delay at times.

    # PRARIT: I've tested this with .25 seconds, and it doesn't make a
    # difference.  The test run takes 2700 vs 2600 seconds to run.
    time.sleep(.5)


def removecomments(intext: str) -> str:
    # remove all lines beginning with # (hash)
    outtext = re.sub(r"^#.*\n", "", intext, flags=re.MULTILINE)
    return outtext


def isGitRepo() -> bool:
    try:
        with open(os.devnull, "w") as devnull:
            subprocess.check_call(
                ["git", "-C", "./", "rev-parse", "--is-inside-work-tree"],
                stdout=devnull,
                stderr=devnull,
            )
        return True
    except subprocess.CalledProcessError:
        return False


def geteditor() -> str:
    editor = os.environ.get("GIT_EDITOR") or os.environ.get("EDITOR") or "vi"
    if not editor:
        print("Could not determine editor.  Please set GIT_EDITOR or EDITOR.")
        sys.exit(1)

    return editor


def editFile(fileprefix: str, message: Optional[str]) -> str:
    editor = geteditor()
    command = shlex.split(editor)

    if isGitRepo():
        workingdir = os.getcwd()
    else:
        workingdir = "/tmp"

    filename = workingdir + "/" + f"{fileprefix}_EDITMSG"

    command.append(filename)

    # prepopulate the file with message
    if message:
        with open(filename, "w") as file:
            file.write(message)

    # open the editor and save the contents in $filename
    try:
        subprocess.run(command, check=True)
    except subprocess.CalledProcessError as e:
        print(f"Editor open failed with: {e}")
        os.remove(filename)
        sys.exit(1)

    # read the saved contents of $filename
    with open(filename, "r") as file:
        intext = file.read()

    # cleanup
    os.remove(filename)

    intext = removecomments(intext)
    intext = re.sub(r"^#.*\n", "", intext, flags=re.MULTILINE)

    return intext


@overload
def dump_any(field: Resource, value: None) -> str: ...


@overload
def dump_any(field: Resource, value: Union[str, int, float, bool]) -> str: ...


@overload
def dump_any(field: Resource, value: Any) -> str: ...


def dump_any(field: Resource, value: Any) -> str:
    if value is None:
        return ""
    # customfield_12316840/"Bugzilla Bug"
    if field.id == "customfield_12316840" and hasattr(value, "bugid"):
        bug = cast(BugzillaBug, value)
        return str(bug.bugid)
    if isinstance(value, (str, int, float, bool)):
        return str(value)
    if hasattr(value, "name"):
        named_resource = cast(JiraNamedResource, value)
        return str(named_resource.name)
    return str(value)


def dump_user(field: Resource, user: User) -> str:
    return f"{user.displayName} <{user.emailAddress}>"


def dump_version(field: Resource, version: Version) -> str:
    return str(version.name)


def dump_component(field: Resource, component: Component) -> str:
    return str(component.name)


def dump_issuelink(field: Resource, issuelink: IssueLink) -> str:
    if hasattr(issuelink, "outwardIssue"):
        return f"{issuelink.type.outward} https://issues.redhat.com/browse/{issuelink.outwardIssue.key}"
    return f"{issuelink.type.inward} https://issues.redhat.com/browse/{issuelink.inwardIssue.key}"


def dump_attachment(field: Resource, attachment: Attachment) -> str:
    return str(attachment.content)


def dump_group(field: Resource, group: Group) -> str:
    return str(group.name)


def dump_array(field: Resource, array: Sequence[Any]) -> str:
    # customfield_12323140/"Target Version"
    if field.id == "customfield_12323140":
        return dump_version(field, cast(Version, array))
    # customfield_12315950/"Contributors"
    if field.id == "customfield_12315950":
        userstr = ""
        count = 0
        for cf in array:
            count += 1
            user = cast(User, cf)
            userstr = userstr + dump_user(field, user)
            if count != len(array):
                userstr = userstr + ", "
        return userstr

    if not array:
        return ""

    if field.schema["items"] and field.schema["items"] != "worklog":
        retstr = ""
        count = 0
        for entry in array:
            count += 1
            match field.schema["items"]:
                case "attachment":
                    retstr = retstr + dump_attachment(field, cast(Attachment, entry))
                case "component":
                    retstr = retstr + dump_component(field, cast(Component, entry))
                case "group":
                    retstr = retstr + dump_group(field, cast(Group, entry))
                case "issuelinks":
                    retstr = retstr + dump_issuelink(field, cast(IssueLink, entry))
                case "option":
                    retstr = retstr + dump_option(field, cast(JiraOption, entry))
                case "string":
                    retstr = retstr + str(entry)
                case "user":
                    retstr = retstr + dump_user(field, cast(User, entry))
                case "version":
                    retstr = retstr + dump_version(field, cast(Version, entry))
                case "worklog":
                    return ""
                case _:
                    print(f"PRARIT unhandled array {field.schema['items']}")
                    return ""
            if count != len(array):
                retstr = retstr + ", "
        return retstr

    return ""


def dump_securitylevel(field: Resource, security: SecurityLevel) -> str:
    return str(security.description)


def dump_option(field: Resource, option: JiraOption) -> str:
    return str(option.value)


def dump_optionwithchild(field: Resource, option: JiraOption) -> str:
    if hasattr(option, "child") and option.child:
        return f"{option.value} - {option.child.value}"
    return str(option.value)


def dump_votes(field: Resource, votes: Votes) -> int:
    return int(votes.votes)


def dump_progress(field: Resource, progress: JiraProgress) -> str:
    return f"{progress.progress}%"


def dump_watches(field: Resource, watches: Watchers) -> str:
    if not watches.isWatching:
        return "0"
    return str(watches.watchCount)


def dump_comment(field: Resource, comment: Comment) -> str:
    creator = dump_user(field, comment.author)
    timestamp = convert_jira_date(comment.created)
    return f'"Created by {creator} at {timestamp} :\\n{comment.body}\\n\\n"'


def dump_comments(field: Resource, comments: List[Comment]) -> str:
    retstr = ""
    count = 0
    for comment in comments:
        count += 1
        retstr = retstr + dump_comment(field, comment)
        if count != len(comments):
            retstr = retstr + ", "
    return retstr


def convert_jira_date(datestr: str) -> str:
    try:
        # 2024-09-03 11:34:05
        date = datetime.strptime(datestr, "%Y-%m-%dT%H:%M:%S.%f%z")
    except Exception:
        # 2025-04-30
        try:
            date = datetime.strptime(datestr, "%Y-%m-%d")
        except Exception:
            print(f"ERROR: undefined date format {datestr}")
            sys.exit(1)
    return date.strftime("%Y-%m-%d %H:%M:%S")


def dict_to_struct(data: Dict[str, Any]) -> Resource:
    return cast(Resource, type("", (object,), data)())


def evaluatefield(field: Resource, value: Any, noescapedtext: bool) -> Optional[Union[str, int]]:
    schema = dict_to_struct(field.schema)
    match schema.type:
        case "any":
            # "Git Pull Request"
            if field.id == "customfield_12310220":
                retstr = ""
                count = 0
                for v in value:
                    count += 1
                    retstr = retstr + str(v)
                    if count != len(value):
                        retstr += ", "
                return retstr
            return str(dump_any(field, value))
        case "array":
            if value is None:
                return ""
            return str(dump_array(field, value))
        case "date":
            if value is None:
                return ""
            return str(convert_jira_date(value))
        case "datetime":
            if value is None:
                return ""
            date = datetime.strptime(value, "%Y-%m-%dT%H:%M:%S.%f%z")
            return date.strftime("%Y-%m-%d %H:%M:%S")
        case "issuelinks":
            if value is None:
                return ""
            return str(dump_issuelink(field, cast(IssueLink, value)))
        case "issuetype":
            return str(value)
        case "number":
            if value is None:
                return "0"
            return int(float(value))
        case "sd-approvals":
            return ""
        case "sd-customerrequesttype":
            return ""
        case "sd-servicelevelagreement":
            return ""
        case "securitylevel":
            if value is None:
                return ""
            return str(dump_securitylevel(field, cast(SecurityLevel, value)))
        case "string":
            if value is None:
                return ""
            if noescapedtext:
                return str(value)
            return str(value).replace("\n", "\\n")
        case "timetracking":
            return ""
        case "option":
            if value is None:
                return ""
            return str(dump_option(field, cast(JiraOption, value)))
        case "option-with-child":
            if value is None:
                return ""
            return str(dump_optionwithchild(field, cast(JiraOption, value)))
        case "priority":
            if value is None:
                return ""
            return str(value)
        case "project":
            if value is None:
                return ""
            return str(value)
        case "progress":
            if value is None:
                return ""
            return str(dump_progress(field, cast(JiraProgress, value)))
        case "resolution":
            if value is None:
                return ""
            return str(value)
        case "status":
            if value is None:
                return ""
            return str(value)
        case "user":
            if value is None:
                return ""
            return str(dump_user(field, cast(User, value)))
        case "version":
            if value is None:
                return ""
            return str(dump_version(field, cast(Version, value)))
        case "votes":
            if value is None:
                return ""
            return int(dump_votes(field, cast(Votes, value)))
        case "watches":
            if value is None:
                return ""
            return str(dump_watches(field, cast(Watchers, value)))
        case "comments-page":
            return str(dump_comments(field, value.comments))
        case _:
            print(
                f"ERROR undefined field type FIELD[{field.name}|{field.id}]: ",
                schema.type,
                value,
            )
            return None


def getfieldlist(
    jirafields: List[Dict[str, Any]], userfields: Optional[List[str]]
) -> List[Resource]:
    # generate a list of fields
    fields: List[Resource] = []
    if not userfields:
        for fielddict in jirafields:
            field = dict_to_struct(fielddict)
            fields.append(field)
        return fields

    for userfield in userfields:
        for fielddict in jirafields:
            field = dict_to_struct(fielddict)
            if field.id == userfield or field.name == userfield:
                fields.append(field)
                continue
    return fields


MAX_RETRIES = 5


class RHJiraFetchError(Exception):
    pass


def _get_issue(jira: JIRA, ticketID: str) -> Issue:
    for attempt in range(1, MAX_RETRIES):
        try:
            issue = jira.issue(ticketID)
            rhjiratax()
            return issue
        except JIRAError as e:
            if attempt == (MAX_RETRIES - 1):
                raise RHJiraFetchError(f"JIRAError {e.status_code}: {e.text}") from e
    raise RHJiraFetchError("Maximum retries exceeded")


def getissue(jira: JIRA, ticketID: str) -> Issue:
    return cast(Issue, retry_jira_operation(_get_issue, jira, ticketID))


def _get_fields(jira: JIRA) -> List[Dict[str, Any]]:
    for attempt in range(1, MAX_RETRIES):
        try:
            fields = jira.fields()
            rhjiratax()
            return fields
        except JIRAError as e:
            if attempt == (MAX_RETRIES - 1):
                raise RHJiraFetchError(f"JIRAError {e.status_code}: {e.text}") from e
    raise RHJiraFetchError("Maximum retries exceeded")


def getfields(jira: JIRA) -> List[Dict[str, Any]]:
    return cast(List[Dict[str, Any]], retry_jira_operation(_get_fields, jira))


def _add_comment(jira: JIRA, ticketID: str, savedText: str) -> None:
    for attempt in range(1, MAX_RETRIES):
        try:
            jira.add_comment(ticketID, savedText)
            rhjiratax()
            return
        except JIRAError as e:
            if attempt == (MAX_RETRIES - 1):
                raise RHJiraFetchError(f"JIRAError {e.status_code}: {e.text}") from e
    raise RHJiraFetchError("Maximum retries exceeded")


def addcomment(jira: JIRA, ticketID: str, savedText: str) -> None:
    retry_jira_operation(_add_comment, jira, ticketID, savedText)


def _create_issue(jira: JIRA, fields: Dict[str, Any]) -> Issue:
    for attempt in range(1, MAX_RETRIES):
        try:
            issue = jira.create_issue(fields=fields)
            rhjiratax()
            return issue
        except JIRAError as e:
            if attempt == (MAX_RETRIES - 1):
                raise RHJiraFetchError(f"JIRAError {e.status_code}: {e.text}") from e
    raise RHJiraFetchError("Maximum retries exceeded")


def createissue(jira: JIRA, fields: Dict[str, Any]) -> Issue:
    return cast(Issue, retry_jira_operation(_create_issue, jira, fields))


def _get_transitions(jira: JIRA, issue: Issue) -> List[Dict[str, Any]]:
    for attempt in range(1, MAX_RETRIES):
        try:
            transitions = jira.transitions(issue)
            rhjiratax()
            return list(transitions)
        except JIRAError as e:
            if attempt == (MAX_RETRIES - 1):
                raise RHJiraFetchError(f"JIRAError {e.status_code}: {e.text}") from e
    raise RHJiraFetchError("Maximum retries exceeded")


def gettransitions(jira: JIRA, issue: Issue) -> List[Dict[str, Any]]:
    return cast(List[Dict[str, Any]], retry_jira_operation(_get_transitions, jira, issue))


def _transition_issue(jira: JIRA, issue: Issue, stateID: str) -> None:
    for attempt in range(1, MAX_RETRIES):
        try:
            jira.transition_issue(issue, stateID)
            rhjiratax()
            return
        except JIRAError as e:
            if attempt == (MAX_RETRIES - 1):
                raise RHJiraFetchError(f"JIRAError {e.status_code}: {e.text}") from e
    raise RHJiraFetchError("Maximum retries exceeded")


def transitionissue(jira: JIRA, issue: Issue, stateID: str) -> None:
    retry_jira_operation(_transition_issue, jira, issue, stateID)


def _assign_issue(jira: JIRA, issue: Issue, assignee: Optional[str]) -> None:
    for attempt in range(1, MAX_RETRIES):
        try:
            jira.assign_issue(issue, assignee)
            rhjiratax()
            return
        except JIRAError as e:
            if attempt == (MAX_RETRIES - 1):
                raise RHJiraFetchError(f"JIRAError {e.status_code}: {e.text}") from e
    raise RHJiraFetchError("Maximum retries exceeded")


def assignissue(jira: JIRA, issue: Issue, assignee: Optional[str]) -> None:
    retry_jira_operation(_assign_issue, jira, issue, assignee)


def _search_issues(jira: JIRA, searchstring: str, maxentries: int) -> List[Issue]:
    for attempt in range(1, MAX_RETRIES):
        try:
            issues = jira.search_issues(searchstring, maxResults=maxentries)
            rhjiratax()
            return cast(List[Issue], list(issues))
        except JIRAError as e:
            if attempt == (MAX_RETRIES - 1):
                raise RHJiraFetchError(f"JIRAError {e.status_code}: {e.text}") from e
    raise RHJiraFetchError("Maximum retries exceeded")


def searchissues(jira: JIRA, searchstring: str, maxentries: int) -> List[Issue]:
    return cast(List[Issue], retry_jira_operation(_search_issues, jira, searchstring, maxentries))


def _create_issue_link(jira: JIRA, linktype: str, inwardissue: str, outwardissue: str) -> None:
    for attempt in range(1, MAX_RETRIES):
        try:
            jira.create_issue_link(
                type=linktype,
                inwardIssue=inwardissue,
                outwardIssue=outwardissue
            )
            rhjiratax()
            return
        except JIRAError as e:
            if attempt == (MAX_RETRIES - 1):
                raise RHJiraFetchError(f"JIRAError {e.status_code}: {e.text}") from e
    raise RHJiraFetchError("Maximum retries exceeded")


def createissuelink(jira: JIRA, linktype: str, inwardissue: str, outwardissue: str) -> None:
    retry_jira_operation(_create_issue_link, jira, linktype, inwardissue, outwardissue)


def handle_jira_error(e: Exception, context: str, additional_info: Optional[str] = None) -> bool:
    if isinstance(e.__cause__, JIRAError):
        jira_error = e.__cause__
        try:
            resp = json.loads(jira_error.response.text)
            print(f"Failed to {context}: {jira_error.status_code}")
            if "errorMessages" in resp:
                for msg in resp["errorMessages"]:
                    print(f"Error: {msg}")
            if "errors" in resp:
                for field, error in resp["errors"].items():
                    print(f"     Field '{field}': {error}")
                    if field == "components" and "is not valid" in str(error):
                        print(
                            f"     Run 'rhjira info --project {additional_info} components' to see valid component names" # noqa: E501
                        )
                    if field == "versions" and "is not valid" in str(error):
                        print(
                            f"     Run 'rhjira info --project {additional_info} versions' to see valid version names" # noqa: E501
                        )
                    if field == "fixVersions" and "is not valid" in str(error):
                        print(
                            f"     Run 'rhjira info --project {additional_info} versions' to see valid version names" # noqa: E501
                        )
            if additional_info:
                print(additional_info)
            # Return True if this was a 401 authentication error
            return jira_error.status_code == 401
        except json.JSONDecodeError:
            print(f"Failed to {context}: {jira_error.status_code} {jira_error.text}")
            return jira_error.status_code == 401
    else:
        print(f"Failed to {context}: {e}")
        return False


def retry_jira_operation(operation: Callable[..., Any], *args: Any, **kwargs: Any) -> Any:
    """
    Retry a JIRA operation with authentication retry on 401 errors.

    Args:
        operation: The JIRA operation function to retry
        *args: Arguments to pass to the operation
        **kwargs: Keyword arguments to pass to the operation

    Returns:
        The result of the operation if successful

    Raises:
        Exception: If the operation fails after all retries
    """
    max_retries = 3
    retry_count = 0

    while retry_count < max_retries:
        try:
            return operation(*args, **kwargs)
        except Exception as e:
            # Check if this was an authentication error
            if handle_jira_error(e, "perform JIRA operation"):
                retry_count += 1
                if retry_count < max_retries:
                    print(f"Authentication failed, retrying ({retry_count}/{max_retries})...")
                    # Re-authenticate by creating a new JIRA client
                    new_jira = login.login()

                    # Replace the old client with the new one in the args tuple
                    args = (new_jira,) + args[1:]

                    # Update the global jira reference in the calling module
                    import inspect
                    frame = inspect.currentframe()
                    if frame and frame.f_back:
                        caller_frame = frame.f_back
                        if 'jira' in caller_frame.f_globals:
                            caller_frame.f_globals['jira'] = new_jira

                    continue
            # If not an auth error or we're out of retries, re-raise
            raise


def format_link(key: str, summary: str, status: str, status_category: str) -> str:
    """Format a JIRA issue link with color-coded status."""
    # Determine color based on status category
    if not status_category:
        color = "\033[34m"  # Default to blue if no category
    elif status_category == "To Do":
        color = "\033[90m"  # Light grey
    elif status_category == "Done":
        color = "\033[32m"  # Green
    else:
        color = "\033[34m"  # Blue

    return f'\033]8;;https://issues.redhat.com/browse/{key}\a{key}\033]8;;\a\033[0;37m {summary} ({color}{status}\033[0m)' # noqa: E501



import argparse
import sys

from jira import JIRA

from rhjira import util


def comment(jira: JIRA) -> None:
    # handle arguments
    sys.argv.remove("comment")
    parser = argparse.ArgumentParser(description="Comment on an RH jira ticket")
    parser.add_argument("-f", "--file", type=str, help="file containing comment")
    parser.add_argument(
        "--noeditor",
        action="store_true",
        help="when set the editor will not be invoked",
    )
    args, ticketIDs = parser.parse_known_args()

    if len(ticketIDs) != 1:
        print(f"Error: ticketID not clear or found: {ticketIDs}")
        sys.exit(1)
    ticketID = ticketIDs[0]

    filename = args.file

    try:
        util.getissue(jira, ticketID)
    except Exception as e:
        util.handle_jira_error(e, f"lookup ticket {ticketID}")
        sys.exit(1)

    editorText = "# Lines beginning with '#' will be ignored"
    if filename:
        # read the saved contents of $filename
        try:
            with open(filename, "r") as file:
                editorText = file.read()
        except Exception as error:
            print(f"Unable to open {args.template}: {error}")
            sys.exit(1)

    savedText = editorText
    if not args.noeditor:
        savedText = util.editFile("rhjira", editorText)
        if len(savedText) == 0:
            print("Empty comment buffer ... aborting.")
            sys.exit(1)

    try:
        util.addcomment(jira, ticketID, savedText)
    except Exception as e:
        util.handle_jira_error(e, f"add comment to {ticketID}")
        print("")
        print("Comment text:")
        print(savedText)
        sys.exit(1)

    print(f"https://issues.redhat.com/browse/{ticketID}")



import argparse
import re
import sys
from typing import Any, Dict, List, Match

from jira import JIRA
from tabulate import tabulate

from rhjira import util


def list(jira: JIRA) -> None:
    # handle arguments
    sys.argv.remove("list")
    parser = argparse.ArgumentParser(description="list issues")
    parser.add_argument(
        "--fields", type=str, help="specify a comma-separated list of fields for output"
    )
    parser.add_argument(
        "--numentries", type=int, help="maximum number of entries to return"
    )
    parser.add_argument(
        "--summarylength",
        type=int,
        help="Length of the summary string (default:50, do not truncate:0)",
    )
    parser.add_argument(
        "--noheader", action="store_true", help="do not display an output header"
    )
    parser.add_argument(
        "--nolinenumber", action="store_true", help="do not display leading line number"
    )
    parser.add_argument(
        "--rawoutput",
        action="store_true",
        help="do not make output pretty (useful for scripts)",
    )
    parser.add_argument(
        "--textonly",
        action="store_true",
        help="do not output URLs",
    )
    args, jqls = parser.parse_known_args()

    if len(jqls) != 1:
        print("Error: No search string provided.")
        sys.exit(1)
    jqlstring = jqls[0]

    try:
        jirafields = util.getfields(jira)
    except Exception as e:
        util.handle_jira_error(e, "lookup fields")
        sys.exit(1)

    # replace 'customfield' with actual name (for example, Constributors)
    lookup: Dict[str, str] = {item["id"]: item["name"] for item in jirafields}

    def replace_cf(match: Match[str]) -> str:
        key = match.group(0)
        return lookup.get(key, key)

    searchstring = re.sub(r"\bcustomfield_\d+\b", replace_cf, jqlstring)

    maxentries = args.numentries if args.numentries else 50

    try:
        issues = util.searchissues(jira, searchstring, maxentries)
        util.rhjiratax()
    except Exception as e:
        util.handle_jira_error(e, "search")
        sys.exit(1)

    if not issues:  # This handles both None and empty list cases
        print("No issues found.")
        sys.exit(0)

    # default is to show the issue key and the summary
    userfields: List[str] = ["key", "summary"]
    if args.fields:
        userfields = args.fields.split(",")
        # issue key is always output
        if "key" not in userfields:
            userfields.insert(0, "key")

    # generate a list of fields
    fields = util.getfieldlist(jirafields, userfields)

    data: List[List[Any]] = []
    count = 0
    issuekeyindex = 0  # used in table output
    for issue in issues:
        entry: List[Any] = []

        if not args.nolinenumber:
            issuekeyindex = 1
            entry.append(count)

        entry.append(issue.key)

        for field in fields:
            try:
                value = getattr(issue.fields, field.id)
                if field.custom and value is None:
                    entry.append("---")
                    continue
            except Exception:
                entry.append("---" if field.custom else "")
                continue

            # this seems strange, but I want the evaluatefield outside of
            # a try/except.  Otherwise, errors in evaluatefield get 'caught'
            entry.append(util.evaluatefield(field, value, False))

        data.append(entry)
        count += 1

    # issue.key is displayed as "Issue" in the first column
    userfields[0] = "Issue"

    if not args.nolinenumber:
        userfields.insert(0, "#")

    if args.rawoutput:
        for entry in data:
            print(*entry, sep="|")
        return

    if args.noheader:
        print(tabulate(data, tablefmt="plain"))
    else:
        tableoutput = tabulate(data, headers=userfields, tablefmt="plain")

        if args.textonly:
            print(tableoutput)
            sys.exit(1)

        # tabulate uses the full string length to calculate column widths.
        # This causes weird table output when using the full URL length (see
        # below).  Instead, just search/replace the issuekey.
        for line in tableoutput.splitlines():
            issuekey = line.split()[issuekeyindex]
            print(
                line.replace(
                    issuekey,
                    f"\033]8;;https://issues.redhat.com/browse/{issuekey}\a{issuekey}\033]8;;\a\033[0;37m",
                )
            )



import sys
from typing import Any

from jira import JIRA, JIRAError

from rhjira import util


def get_project_info(jira: JIRA, project_key: str, info_type: str) -> list[dict[str, Any]]:
    try:
        if info_type == "components":
            components = jira.project_components(project_key)
            return [{"name": c.name} for c in components]
        else:  # versions
            versions = jira.project_versions(project_key)
            return [{"name": v.name} for v in versions]
    except JIRAError as e:
        util.handle_jira_error(e, f"fetch {info_type} for project {project_key}")
        sys.exit(1)


def display_components(components: list[dict[str, Any]]) -> None:
    if components:
        for c in components:
            print(c["name"])
    else:
        print("No components found.")


def display_versions(versions: list[dict[str, Any]]) -> None:
    if versions:
        for v in versions:
            print(v["name"])
    else:
        print("No versions found.")


def info(jira: JIRA) -> None:
    # handle arguments
    sys.argv.remove("info")

    if len(sys.argv) != 4 or sys.argv[1] != "--project":
        print("Usage: rhjira info --project <project> <components|versions>")
        sys.exit(1)

    project = sys.argv[2]
    info_type = sys.argv[3]

    if info_type not in ["components", "versions"]:
        print("Error: info type must be either 'components' or 'versions'")
        sys.exit(1)

    info_list = get_project_info(jira, project, info_type)

    if info_type == "components":
        display_components(info_list)
    else:
        display_versions(info_list)



import argparse
import sys
from typing import Optional, Union

from jira import JIRA
from jira.resources import Resource

from rhjira import util

showcustomfields = True


def outputfield(field: Resource, data: Optional[Union[str, int]]) -> None:
    if showcustomfields:
        if data is None or data == "":
            print(f"FIELD[{field.name}|{field.id}]:")
        else:
            print(f"FIELD[{field.name}|{field.id}]: {data}")
    else:
        if data is None or data == "":
            print(f"FIELD[{field.name}]:")
        else:
            print(f"FIELD[{field.name}]: {data}")


noescapedtext = True


def dump(jira: JIRA) -> None:
    # handle arguments
    sys.argv.remove("dump")
    parser = argparse.ArgumentParser(description="Dump jira issue variables.")
    parser.add_argument(
        "--fields", type=str, help="specify a comma-separated list of fields for output"
    )
    parser.add_argument(
        "--noescapedtext",
        action="store_true",
        help="show dump text (ie, no escaped characters)",
    )
    parser.add_argument(
        "--showcustomfields",
        action="store_true",
        help="show the customfield IDs and the customfield names",
    )
    parser.add_argument(
        "--showemptyfields",
        action="store_true",
        help="show all fields including those that are not defined for this issue",
    )
    parser.add_argument("--json", action="store_true", help="dump issue in json format")
    args, ticketIDs = parser.parse_known_args()

    if len(ticketIDs) != 1:
        print(f"Error: ticketID not clear or found: {ticketIDs}")
        sys.exit(1)
    ticketID = ticketIDs[0]

    userfields = args.fields
    if args.fields and len(args.fields) != 0:
        userfields = args.fields.split(",")
    global noescapedtext
    noescapedtext = args.noescapedtext
    global showcustomfields
    showcustomfields = args.showcustomfields
    showemptyfields = args.showemptyfields

    try:
        issue = util.getissue(jira, ticketID)
    except Exception as e:
        util.handle_jira_error(e, f"lookup ticket {ticketID}")
        sys.exit(1)

    if args.json:
        util.dumpissue(issue)
        sys.exit(0)

    try:
        jirafields = util.getfields(jira)
    except Exception as e:
        util.handle_jira_error(e, "lookup fields")
        sys.exit(1)

    fields = util.getfieldlist(jirafields, userfields)

    # output the fields
    for field in fields:
        if field.id in issue.raw["fields"]:
            if issue.raw["fields"][field.id]:
                try:
                    value = getattr(issue.fields, field.id)
                except AttributeError:
                    value = ""

                outputfield(field, util.evaluatefield(field, value, noescapedtext))
            else:
                if showemptyfields:
                    outputfield(field, "")
        else:
            # For some reason the issue key field is not populated with a value.
            if field.id == "issuekey":
                outputfield(field, ticketID)
                continue
            if showemptyfields:
                outputfield(field, "")



import json
from unittest import TestCase
from unittest.mock import MagicMock, patch

from jira import JIRA, JIRAError

from rhjira.util import handle_jira_error, retry_jira_operation


class TestRetryJiraOperation(TestCase):
    def setUp(self) -> None:
        self.mock_jira = MagicMock(spec=JIRA)
        self.mock_operation = MagicMock()

    def test_successful_operation(self) -> None:
        """Test that a successful operation returns the expected result."""
        expected_result = "success"
        self.mock_operation.return_value = expected_result

        result = retry_jira_operation(self.mock_operation, self.mock_jira, "arg1", "arg2")

        self.assertEqual(result, expected_result)
        self.mock_operation.assert_called_once_with(self.mock_jira, "arg1", "arg2")

    @patch('rhjira.util.login.login')
    def test_auth_error_with_successful_retry(self, mock_login: MagicMock) -> None:
        """Test that an auth error is handled and retried successfully."""
        # First call fails with auth error, second succeeds
        mock_response = MagicMock()
        mock_response.text = json.dumps({"errorMessages": ["Authentication failed"]})
        jira_error = JIRAError(status_code=401, response=mock_response)
        exception = Exception("Test error")
        exception.__cause__ = jira_error

        self.mock_operation.side_effect = [exception, "success"]
        mock_login.return_value = self.mock_jira

        result = retry_jira_operation(self.mock_operation, self.mock_jira, "arg1", "arg2")

        self.assertEqual(result, "success")
        self.assertEqual(self.mock_operation.call_count, 2)
        mock_login.assert_called_once()

    @patch('rhjira.util.login.login')
    def test_auth_error_max_retries_exceeded(self, mock_login: MagicMock) -> None:
        """Test that max retries are enforced for auth errors."""
        # All calls fail with auth error
        mock_response = MagicMock()
        mock_response.text = json.dumps({"errorMessages": ["Authentication failed"]})
        jira_error = JIRAError(status_code=401, response=mock_response)
        exception = Exception("Test error")
        exception.__cause__ = jira_error

        self.mock_operation.side_effect = exception
        mock_login.return_value = self.mock_jira

        with self.assertRaises(Exception): # noqa B017
            retry_jira_operation(self.mock_operation, self.mock_jira, "arg1", "arg2")

        self.assertEqual(self.mock_operation.call_count, 3)  # Max retries
        self.assertEqual(mock_login.call_count, 2)  # Called twice for retries

    def test_non_auth_error(self) -> None:
        """Test that non-auth errors are not retried."""
        self.mock_operation.side_effect = JIRAError(status_code=404, text="Not found")

        with self.assertRaises(JIRAError):
            retry_jira_operation(self.mock_operation, self.mock_jira, "arg1", "arg2")

        self.mock_operation.assert_called_once()


class TestHandleJiraError(TestCase):
    def setUp(self) -> None:
        self.mock_response = MagicMock()

    def test_jira_error_with_error_messages(self) -> None:
        """Test handling of JIRAError with errorMessages."""
        error_messages = ["Error 1", "Error 2"]
        self.mock_response.text = json.dumps({"errorMessages": error_messages})
        jira_error = JIRAError(status_code=400, response=self.mock_response)
        exception = Exception("Test error")
        exception.__cause__ = jira_error

        result = handle_jira_error(exception, "test operation")

        self.assertFalse(result)  # Not an auth error
        # Note: We can't easily test the print output, but we can verify the function returns False

    def test_jira_error_with_errors(self) -> None:
        """Test handling of JIRAError with errors field."""
        errors = {
            "components": "is not valid",
            "versions": "is not valid",
            "fixVersions": "is not valid"
        }
        self.mock_response.text = json.dumps({"errors": errors})
        jira_error = JIRAError(status_code=400, response=self.mock_response)
        exception = Exception("Test error")
        exception.__cause__ = jira_error

        result = handle_jira_error(exception, "test operation", "TEST-PROJ")

        self.assertFalse(result)  # Not an auth error

    def test_jira_error_json_decode_error(self) -> None:
        """Test handling of JIRAError with invalid JSON response."""
        self.mock_response.text = "invalid json"
        jira_error = JIRAError(status_code=400, response=self.mock_response)
        exception = Exception("Test error")
        exception.__cause__ = jira_error

        result = handle_jira_error(exception, "test operation")

        self.assertFalse(result)  # Not an auth error

    def test_non_jira_error(self) -> None:
        """Test handling of non-JIRAError exception."""
        exception = Exception("Test error")

        result = handle_jira_error(exception, "test operation")

        self.assertFalse(result)  # Not an auth error

    def test_auth_error(self) -> None:
        """Test handling of authentication error (401)."""
        self.mock_response.text = json.dumps({"errorMessages": ["Authentication failed"]})
        jira_error = JIRAError(status_code=401, response=self.mock_response)
        exception = Exception("Test error")
        exception.__cause__ = jira_error

        result = handle_jira_error(exception, "test operation")

        self.assertTrue(result)  # Is an auth error



import argparse
from datetime import datetime
import os
import sys
from typing import Any, Dict, List, Optional

from jira import JIRA

from rhjira import util


# ruff: noqa: E501
def getnames(data: str) -> Optional[List[Dict[str, str]]]:
    if data.strip() == "":
        return None

    entries = []
    for item in data.split(","):
        entry = {"name": item.strip()}
        entries.append(entry)
    return entries


def create(jira: JIRA) -> None:
    # handle arguments
    sys.argv.remove("create")
    parser = argparse.ArgumentParser(
        description="Create a jira ticket on https://issues.redhat.com"
    )
    parser.add_argument("--affectsversion", type=str, help="Affected Version.")
    parser.add_argument(
        "--assignee", type=str, help="Assignee email address used in RH Jira."
    )
    parser.add_argument(
        "--components", type=str, help="Components (must be specified)."
    )
    parser.add_argument(
        "--contributors",
        type=str,
        help="Contributors (comma separated list of email addresses).",
    )
    parser.add_argument("--description", type=str, help="Description for the ticket.")
    parser.add_argument("--epiclink", type=str, help="Epic Link")
    parser.add_argument("--epicname", type=str, help="Epic Name")
    parser.add_argument("--fixversion", type=str, help="Fix/Versions.")
    parser.add_argument(
        "--noeditor",
        action="store_true",
        help="when set the editor will not be invoked",
    )
    parser.add_argument("--parentlink", type=str, help="Parent Link")
    parser.add_argument("--priority", type=str, help="Priority, Normal by default.")
    parser.add_argument(
        "--project", type=str, help="Project Name (RHEL, RHELAI, AIPCC, etc.)."
    )
    parser.add_argument("--releaseblocker", type=str, help="Release Blocker Status.")
    parser.add_argument("--severity", type=str, help="Severity Level.")
    parser.add_argument(
        "--summary",
        type=str,
        help="Summary for the ticket.  Cannot exceed more than one line.",
    )
    parser.add_argument(
        "-T",
        "--template",
        type=str,
        help="Create jira with template file (will be opened in $GITLAB_EDITOR)",
    )
    parser.add_argument(
        "--tickettype", type=str, help="Ticket Type (Epic, Story, Bug, etc.)."
    )
    parser.add_argument(
        "--isblockedby", type=str, help="issue which blocks this new issue"
    )
    parser.add_argument(
        "--blocks", type=str, help="issue that is blocked by this new issue"
    )
    args = parser.parse_args()

    # clear unset vars
    for arg in vars(args):
        if getattr(args, arg) is None:
            setattr(args, arg, "")

    # The description is a special case.  If it's been passed on the command line with a \n,
    # then we need to parse this so it looks like real text.
    description = util.convertdescription(args.description)

    defaultCreateText = f"""# Please fill out the following fields.  This file can be saved to create a template.
# Project (RHEL, RHELAI, AIPCC, RHELENG, etc.)
Project: {args.project}
# Ticket Type (Epic, Bug, Story, etc.)
Ticket Type: {args.tickettype}
# Epic Name (required if Ticket Type is Epic, o/w ignored)
Epic Name: {args.epicname}
# Summary (one line only)
Summary: {args.summary}
# Description (multi line accepted)
# Note: all 'free' lines in text file will be accepted as part of the Description
Description: {description}
# Assignee (email address)
Assignee: {args.assignee}
# Components (case sensitive)
Components: {args.components}
# Affects Versions
Affects Versions: {args.affectsversion}
# Fix Version
Fix Version: {args.fixversion}
# Priority
Priority: {args.priority}
#
# The fields below may not be available on all projects.
#
# Contributors (comma separated list of email addresses)
Contributors: {args.contributors}
# Release Blocker (default to None)
Release Blocker: {args.releaseblocker}
# Severity
Severity: {args.severity}
# Epic Link
Epic Link: {args.epiclink}
# Parent Link
Parent Link: {args.parentlink}
# Is Blocked By
Is Blocked By: {args.isblockedby}
# Blocks
Blocks: {args.blocks}
"""

    createtext = defaultCreateText
    if args.template != "":
        try:
            with open(args.template, "r") as file:
                createtext = file.read()
        except Exception as error:
            print(f"Unable to open {args.template}: {error}")
            sys.exit(1)

    if not args.noeditor:
        createtext = util.editFile("rhjira", createtext)

    if len(createtext) == 0:
        print("No text found ... aborting.")
        sys.exit(1)

    fields: Dict[str, Any] = {}
    description = ""
    isblockedby = ""
    blocks = ""
    for line in createtext.splitlines():
        token = line.split(":")[0] + ":"
        data = line[len(token) + 1 :].strip()

        match token:
            case "Project:":
                fields["project"] = {"key": data.strip()}
            case "Ticket Type:":
                tickettype = data.strip()
                fields["issuetype"] = {"name": data.strip()}
            case "Epic Name:":
                if tickettype == "Epic":
                    fields["customfield_12311141"] = data.strip()
            case "Summary:":
                fields["summary"] = data.strip()
            case "Description:":
                description = data.strip()
            case "Assignee:":
                if data != "":
                    fields["assignee"] = {"name": data}
            case "Components:":
                if data != "":
                    fields["components"] = getnames(data)
            case "Affects Versions:":
                if data != "":
                    fields["versions"] = getnames(data)
            case "Fix Version:":
                if data != "":
                    fields["fixVersions"] = getnames(data)
            case "Priority:":
                if data != "":
                    fields["priority"] = {"name": data.strip()}
            case "Contributors:":
                if data != "":
                    fields["customfield_12315950"] = getnames(data)
            case "Release Blocker:":
                if data != "":
                    fields["customfield_12319743"] = {"value": data}
            case "Severity:":
                if data != "":
                    fields["customfield_12316142"] = {"value": data}
            case "Epic Link:":
                if data != "" and tickettype != "Epic":
                    fields["customfield_12311140"] = data.strip()
            case "Parent Link:":
                if data != "" and tickettype == "Epic":
                    fields["customfield_12313140"] = data.strip()
            case "Is Blocked By:":
                if data != "":
                    isblockedby = data.strip()
            case "Blocks:":
                if data != "":
                    blocks = data.strip()
            case _:
                description = description + "\n" + line

    fields["description"] = description

    try:
        issue = util.createissue(jira, fields)
        # Create issue links if specified
        if isblockedby:
            util.createissuelink(
                jira,
                'Blocks',  # Link type is always "Blocks"
                isblockedby,  # The blocking issue is the inward issue
                issue.key  # The blocked issue is the outward issue
            )
        if blocks:
            util.createissuelink(
                jira,
                'Blocks',  # Link type is always "Blocks"
                issue.key,  # The blocking issue is the inward issue
                blocks  # The blocked issue is the outward issue
            )
        print(f"https://issues.redhat.com/browse/{issue.key}")
    except Exception as e:
        util.handle_jira_error(e, "create issue", fields["project"]["key"])

        tmpdir = os.environ.get("TMPDIR")
        if tmpdir is None:
            tmpdir = "/tmp"
        timestamp = datetime.now().strftime("%Y-%m-%d_%H-%M-%S")
        try:
            with open(f"{tmpdir}/rhjira.{timestamp}", "w") as file:
                file.write(createtext)
                file.close()
        except Exception as error:
            print(f"Unable to save error template file to {tmpdir}/rhjira.{timestamp}: {error}")
            sys.exit(1)
        filepath = f"{tmpdir}/rhjira.{timestamp}"

        print(
            f"To re-run the command with the existing options execute 'rhjira create -T {filepath}'"
        )
        sys.exit(1)



"""Display JIRA ticket hierarchy."""

import argparse
import sys

from jira import Issue, JIRA

from rhjira import util


def display_hierarchy(jira: JIRA, issue: Issue, indent: int = 0) -> None:
    """Display the hierarchy of a JIRA ticket, including child issues and blocking relationships."""
    # Format the current issue
    indent_str = "        " * indent
    if indent == 0:  # Only show the full issue info for the root issue
        print(f"{indent_str}{issue.fields.issuetype.name}: {util.format_link(issue.key, issue.fields.summary, issue.fields.status.name, issue.fields.status.statusCategory.name)}")

    # Get blocking relationships
    if hasattr(issue.fields, 'issuelinks'):
        for link in issue.fields.issuelinks:
            if hasattr(link, 'inwardIssue') and link.type.inward == 'is blocked by':
                blocked_by_key = link.inwardIssue.key
                try:
                    blocked_by = util.getissue(jira, blocked_by_key)
                    print(f"{indent_str}        Is Blocked By: {blocked_by.fields.issuetype.name}: {util.format_link(blocked_by.key, blocked_by.fields.summary, blocked_by.fields.status.name, blocked_by.fields.status.statusCategory.name)}")
                    # Recursively display blocking issue's hierarchy
                    display_hierarchy(jira, blocked_by, indent + 1)
                except Exception:
                    # If we can't get the full issue, just show what we have
                    print(f"{indent_str}        Is Blocked By: {util.format_link(blocked_by_key, link.inwardIssue.fields.summary, link.inwardIssue.fields.status.name, link.inwardIssue.fields.status.statusCategory.name)}")

    # Get child issues
    child_issues = util.searchissues(jira, f"'Parent Link' = {issue.key}", 2000)
    for child in child_issues:
        print(f"{indent_str}        {child.fields.issuetype.name}: {util.format_link(child.key, child.fields.summary, child.fields.status.name, child.fields.status.statusCategory.name)}")
        # Recursively display child's hierarchy
        display_hierarchy(jira, child, indent + 1)


def hierarchy(jira: JIRA) -> None:
    """Display the hierarchy of a JIRA ticket."""
    # handle arguments
    sys.argv.remove("hierarchy")
    parser = argparse.ArgumentParser(
        description="Show hierarchy information for a RH jira ticket"
    )
    args, ticketIDs = parser.parse_known_args()

    if len(ticketIDs) != 1:
        print(f"Error: ticketID not clear or found: {ticketIDs}")
        sys.exit(1)
    ticketID = ticketIDs[0]

    try:
        issue = util.getissue(jira, ticketID)
    except Exception as e:
        util.handle_jira_error(e, f"lookup ticket {ticketID}")
        sys.exit(1)

    display_hierarchy(jira, issue)



import getpass
import os
import sys
from typing import Optional

from jira import JIRA, JIRAError
import keyring

from rhjira import util

MAX_RETRIES = 6


def connect2jira(server: str, token: str) -> JIRA:
    last_error: Optional[JIRAError] = None
    for attempt in range(1, MAX_RETRIES):
        try:
            jira = JIRA(server=server, token_auth=token)
            jira.myself()  # test to see if this works
            util.rhjiratax()
            return jira
        except JIRAError as e:
            if attempt == (MAX_RETRIES - 1):
                print(
                    "Error: login failure.  This is likely caused by an invalid token.  Please verify your token and try again." # noqa: E501
                )
                util.handle_jira_error(e, "login")
                sys.exit(1)

    if last_error:
        print("ERROR: unknown login error")
    sys.exit(1)  # Fallback exit if we somehow get here


def login() -> JIRA:
    # check the keyring first
    token: Optional[str] = None
    try:
        token = keyring.get_password("rhjira", getpass.getuser())
    except Exception:
        pass

    if token is None or token == "":
        token = os.getenv("JIRA_TOKEN")
        if token is None or token == "":
            print("Error: The keyring password could not be read and JIRA_TOKEN was not set.")
            sys.exit(1)

    return connect2jira("https://issues.redhat.com", token)


def setpassword() -> None:
    token = getpass.getpass("Enter in token (hit ENTER to abort):")
    if token.strip() == "":
        print("No token entered ... aborting")
        sys.exit(1)

    username = getpass.getuser()
    print(f"Attempting to save token to {username}'s keyring ....")
    keyring.set_password("rhjira", username, token)
    print("testing login with token ....")

    login()

    print(f"Login succeeded.  Token is saved in keyring as ('{username}' on 'rhjira').")



import argparse
import sys
from typing import Any, Dict, List, Optional, Sequence

from jira import Issue, JIRA, JIRAError
from jira.resources import Resource

from rhjira import util


def transitionissue(jira: JIRA, issue: Issue, newstatus: str) -> None:
    if issue.fields.status.name == newstatus:
        print(f"issue {issue.key} is already {newstatus}")
        sys.exit(0)

    try:
        transitions = util.gettransitions(jira, issue)
    except Exception as e:
        util.handle_jira_error(e, f"lookup transitions for {issue.key}")
        sys.exit(1)

    for transition in transitions:
        # according to Jira documentation, the newstatus status must exist.
        if transition["name"] == newstatus:
            try:
                util.transitionissue(jira, issue, transition["id"])
                # note, rhjiratax must be handled on caller side
                return
            except Exception as e:
                util.handle_jira_error(e, f"transition to {newstatus} for {issue.key}")
                sys.exit(1)


def string2entries(data: str) -> Optional[List[Dict[str, str]]]:
    if data.strip() == "":
        return None

    entries = []
    for item in data.split(","):
        entry = {"name": item.strip()}
        entries.append(entry)
    return entries


def entries2string(data: Optional[Sequence[Resource]]) -> str:
    retstr = ""

    if data is None:
        return retstr

    count = 0
    for item in data:
        count += 1
        retstr = retstr + item.name
        if count < len(data):
            retstr = retstr + ", "
    return retstr


def edit(jira: JIRA) -> None:
    # handle arguments
    closejira = False
    if "close" in sys.argv:
        closejira = True
        sys.argv.remove("close")

    if "edit" in sys.argv:
        sys.argv.remove("edit")

    parser = argparse.ArgumentParser(
        description="Create a jira ticket on https://issues.redhat.com"
    )
    parser.add_argument(
        "--affectsversion", type=str, default=None, help="Affected Version."
    )
    parser.add_argument(
        "--assignee",
        type=str,
        default=None,
        help="Assignee email address used in RH Jira.",
    )
    parser.add_argument(
        "--close", action="store_true", help="Close Ticket/Set status to 'Closed'"
    )
    parser.add_argument(
        "--components", type=str, default=None, help="Components (must be specified)."
    )
    parser.add_argument(
        "--contributors",
        type=str,
        default=None,
        help="Contributors (comma separated list of email addresses).",
    )
    parser.add_argument(
        "--description", type=str, default=None, help="Description for the ticket."
    )
    parser.add_argument("--epiclink", type=str, default=None, help="Epic Link")
    parser.add_argument("--epicname", type=str, default=None, help="Epic Name")
    parser.add_argument("--fixversion", type=str, default=None, help="Fix/Versions.")
    parser.add_argument(
        "--gitpullrequest", type=str, default=None, help="Git Pull Request"
    )
    parser.add_argument(
        "--noeditor",
        action="store_true",
        help="when set the editor will not be invoked",
    )
    parser.add_argument("--parentlink", type=str, default=None, help="Parent Link")
    parser.add_argument(
        "--priority", type=str, default=None, help="Priority, Normal by default."
    )
    parser.add_argument(
        "--releaseblocker", type=str, default=None, help="Release Blocker Status."
    )
    parser.add_argument(
        "--status",
        type=str,
        default=None,
        help="Status (New, In Progress, Review, Closed, etc.)",
    )
    parser.add_argument("--severity", type=str, default=None, help="Severity Level.")
    parser.add_argument(
        "--summary",
        type=str,
        default=None,
        help="Summary for the ticket.  Cannot exceed more than one line.",
    )
    parser.add_argument(
        "--isblockedby", type=str, default=None, help="issue which blocks this issue"
    )
    parser.add_argument(
        "--blocks", type=str, default=None, help="issue that is blocked by this issue"
    )
    args, ticketIDs = parser.parse_known_args()

    if len(ticketIDs) != 1:
        print(f"Error: ticketID not clear or found: {ticketIDs}")
        sys.exit(1)
    ticketID = ticketIDs[0]

    try:
        issue = util.getissue(jira, ticketID)
    except Exception as e:
        util.handle_jira_error(e, f"lookup ticket {ticketID}")
        sys.exit(1)

    # if 'close' is specified, just close the Jira and return
    if closejira:
        transitionissue(jira, issue, "Closed")
        print(f"Closed https://issues.redhat.com/browse/{ticketID}")
        sys.exit(0)

    # clear unset vars
    # for arg in vars(args):
    #    if getattr(args, arg) is None:
    #        setattr(args, arg, "")

    epicname = ""
    if issue.fields.issuetype.name == "Epic":
        epicname = issue.fields.customfield_12311141
    if args.epicname is not None:
        epicname = args.epicname

    summary = issue.fields.summary
    if args.summary is not None:
        summary = args.summary

    description = issue.fields.description
    if args.description is not None:
        description = args.description

    assignee = ""
    if issue.fields.assignee is not None:
        assignee = issue.fields.assignee.name
    if args.assignee is not None:
        assignee = args.assignee

    components = entries2string(issue.fields.components)
    if args.components is not None:
        components = args.components

    affectsversion = entries2string(issue.fields.versions)
    if args.affectsversion is not None:
        affectsversion = args.affectsversion

    fixversion = entries2string(issue.fields.fixVersions)
    if args.fixversion is not None:
        fixversion = args.fixversion

    priority = issue.fields.priority.name
    if args.priority is not None:
        priority = args.priority

    contributors = entries2string(issue.fields.customfield_12315950)
    if args.contributors is not None:
        contributors = args.contributors

    releaseblocker = ""
    if issue.fields.customfield_12319743 is not None:
        releaseblocker = issue.fields.customfield_12319743
    if args.releaseblocker is not None:
        releaseblocker = args.releaseblocker

    severity = ""
    if issue.fields.customfield_12316142 is not None:
        severity = issue.fields.customfield_12316142
    if args.severity is not None:
        severity = args.severity

    isblockedby = ""
    if args.isblockedby is not None:
        isblockedby = args.isblockedby

    blocks = ""
    if args.blocks is not None:
        blocks = args.blocks

    epiclink = ""
    if (
        issue.fields.issuetype.name != "Epic"
        and issue.fields.customfield_12311140 is not None
    ):
        epiclink = issue.fields.customfield_12311140
    if args.epiclink is not None:
        epiclink = args.epiclink

    parentlink = ""
    if (
        issue.fields.issuetype.name == "Epic"
        and issue.fields.customfield_12313140 is not None
    ):
        parentlink = issue.fields.customfield_12313140
    if args.parentlink is not None:
        parentlink = args.parentlink

    gitpullrequest = ""
    if issue.fields.issuetype.name in ["Bug", "Story"]:
        if issue.fields.customfield_12310220 is not None:
            gitpullrequest = ",".join(issue.fields.customfield_12310220)
    if args.gitpullrequest is not None:
        gitpullrequest = args.gitpullrequest

    status = issue.fields.status.name
    if args.status is not None:
        status = args.status

    # Due to the complexities of doing so, tickettype cannot be changed in the CLI.
    # This value is referenced while configuring fields.
    issuetype = issue.fields.issuetype.name

    defaultEditText = f"""# Edit {issuetype} {ticketID}. The following fields can be modified.
# Epic Name (required if Ticket Type is Epic, o/w ignored)
Epic Name: {epicname}
# Summary (one line only)
Summary: {summary}
# Status (New, In Progress, Closed, etc.)
Status: {status}
# Description (multi line accepted)
# Note: all 'free' lines in text file will be accepted as part of the Description
Description: {description}
# Assignee (email address)
Assignee: {assignee}
# Components (case sensitive)
Components: {components}
# Affects Versions
Affects Versions: {affectsversion}
# Fix Version
Fix Version: {fixversion}
# Priority
Priority: {priority}
#
# The fields below may not be available on all projects.
#
# Contributors (comma separated list of email addresses)
Contributors: {contributors}
# Release Blocker (default to None)
Release Blocker: {releaseblocker}
# Severity
Severity: {severity}
# Epic Link
Epic Link: {epiclink}
# Parent Link
Parent Link: {parentlink}
# Git Pull Request
Git Pull Request: {gitpullrequest}
# Is Blocked By
Is Blocked By: {isblockedby}
# Blocks
Blocks: {blocks}
"""

    edittext = ""
    if args.noeditor:
        edittext = util.removecomments(defaultEditText)
    else:
        origtext = util.removecomments(defaultEditText)
        edittext = util.editFile("rhjira", defaultEditText)
        if origtext == edittext:
            print("No changes made in editor.... aborting.")
            sys.exit(0)

    fields: Dict[str, Any] = {}
    description = ""
    for line in edittext.splitlines():
        token = line.split(":")[0] + ":"
        data = line[len(token) + 1 :].strip()

        match token:
            case "Project:":
                fields["project"] = {"key": data.strip()}
            case "Ticket Type:":
                fields["issuetype"] = {"name": data.strip()}
            case "Epic Name:":
                if issuetype == "Epic":
                    fields["customfield_12311141"] = data.strip()
            case "Summary:":
                fields["summary"] = data.strip()
            case "Description:":
                description = data.strip()
            case "Assignee:":
                # this requires different permissions than editing so it's a separate
                # jira call
                assignee = data.strip()
                if issue.fields.assignee is not None:
                    if assignee != issue.fields.assignee.name:
                        assignee = data.strip()

                if assignee == "":
                    assignee = None  # type: ignore
                try:
                    util.assignissue(jira, issue, assignee)
                    util.rhjiratax()
                except Exception as e:
                    util.handle_jira_error(e, "assign issue")
                    sys.exit(1)
            case "Components:":
                if data == "":
                    fields["components"] = []
                if data != "":
                    fields["components"] = string2entries(data)
            case "Affects Versions:":
                fields["versions"] = []
                if data != "":
                    fields["versions"] = string2entries(data)
            case "Fix Version:":
                fields["fixVersions"] = []
                if data != "":
                    fields["fixVersions"] = string2entries(data)
            case "Priority:":
                if data != "":
                    fields["priority"] = {"name": data.strip()}
            case "Contributors:":
                fields["customfield_12315950"] = []
                if data != "":
                    fields["customfield_12315950"] = string2entries(data)
            case "Release Blocker:":
                if data == "" and issue.fields.customfield_12319743 is None:
                    continue
                if data != "":
                    fields["customfield_12319743"] = {"value": data}
                else:
                    fields["customfield_12319743"] = None
            case "Severity:":
                if data != "":
                    fields["customfield_12316142"] = {"value": data}
            case "Is Blocked By:":
                if data != "":
                    isblockedby = data.strip()
            case "Blocks:":
                if data != "":
                    blocks = data.strip()
            case "Epic Link:":
                if issuetype in ["Epic", "Feature"]:
                    continue
                if data != "":
                    fields["customfield_12311140"] = data.strip()
                else:
                    fields["customfield_12311140"] = None
            case "Parent Link:":
                if issuetype != "Epic":
                    continue
                fields["customfield_12313140"] = ""
                if data != "":
                    fields["customfield_12313140"] = data.strip()
            case "Git Pull Request:":
                if issuetype in ["Bug", "Story"]:
                    fields["customfield_12310220"] = ""
                    if data != "":
                        fields["customfield_12310220"] = f"{data}"
                    else:
                        fields["customfield_12310220"] = None
            case "Status:":
                status = data.strip()
            case _:
                # this means every 'uncommented' line not in the case statements is
                # part of the description
                description = description + "\n" + line

    # The description is a special case.  If it's been passed on the command line with a \n,
    # then we need to parse this so it looks like real text.
    description = util.convertdescription(description)

    fields["description"] = description

    if status != issue.fields.status.name:
        transitionissue(jira, issue, status)
        util.rhjiratax()

    try:
        issue.update(fields=fields)
        # Create issue links if specified
        if isblockedby:
            util.createissuelink(
                jira,
                'Blocks',  # Link type is always "Blocks"
                isblockedby,  # The blocking issue is the inward issue
                issue.key  # The blocked issue is the outward issue
            )
        if blocks:
            util.createissuelink(
                jira,
                'Blocks',  # Link type is always "Blocks"
                issue.key,  # The blocking issue is the inward issue
                blocks  # The blocked issue is the outward issue
            )
        print(f"https://issues.redhat.com/browse/{ticketID}")
    except JIRAError as e:
        util.handle_jira_error(e, "update issue", fields["project"]["key"])
        sys.exit(1)
